/************************************************************************\
 
 * This file implements the main B&B routine of the TCD algorithm [1]
 
 * Signature:
 R = tcd_test(S1,S2,L,opt);
 * Input:
 S1,S2,S3:      the integral images of vid1 and vid2 respectively (can be
 generated by provided matlab script initSoft.m)
 bseg1, bseg2, bseg3:
 L:          minimal length allowed in TCD (\ell)
 T:          delay time
 opt:
 .maxIter:   max #iter where TCD terminates [default:1e4]
 .nRes:      number of resulting commonalties [default:1]
 .dist:      metrics of interest ("l1","l2","X2","int") [default:"l1"]
 .isVerbose: verbose or not [default:true]
 
 * Output:
 R:          discovered rectangle sets in the temporal search space
 info:       some info output from the tcd algorithm
 
 * Ref: [1] Unsupervised Temporal Commonality Discovery, ECCV 2012.
 
 * Aug-8-2014 updated by Wen-Sheng Chu (wschu@cmu.edu)
 
 \***********************************************************************/

#include "mex.h" /* Always include this */
#include "usd.h"

#define EPS (float)1.192093e-07 // FLT_EPSILON

// Get parameters from inputs
int getParams(const int nrhs, const mxArray *prhs[], short& L, short& T, unsigned int& maxIter, \
              short& nRes, string& dist, bool& isVerbose) {
    
    // Get L and T
    L = (short)mxGetScalar(prhs[6]);
    T = (short)mxGetScalar(prhs[7]);
    
    // Get options
    maxIter   = (unsigned int)10000;
    nRes      = (unsigned int)1;
    dist      = "l1";
    isVerbose = true;
    
    if (nrhs >= 9) {
        // Get option "maxIter"
        mxArray* bufMaxIter = mxGetField(prhs[8],0,"maxIter");
        if ( bufMaxIter != NULL )
            maxIter = (unsigned int)mxGetScalar(bufMaxIter);
        
        // Get option "nRes"
        mxArray* bufRes = mxGetField(prhs[8],0,"nRes");
        if ( bufRes != NULL )
            nRes = (short)mxGetScalar(bufRes);
        
        // Get option "dist"
        mxArray* bufDist = mxGetField(prhs[8],0,"dist");
        if ( bufDist != NULL )
            dist = mxArrayToString(bufDist);
        
        // Get option "isVerbose"
        mxArray* bufVerb = mxGetField(prhs[8],0,"isVerbose");
        if ( bufVerb != NULL )
            isVerbose = (mxGetScalar(bufVerb) == 1.0f);
    }
    
    // Display
    mexPrintf("+ Load parameters: L=%d, T=%d, ", L, T);
    mexPrintf("maxIter=%d, nRes=%d, dist=\"%s\", isVerbose=%d\n", maxIter, nRes, dist.c_str(), isVerbose);
    return 0;
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    
    // Input checks
    if (nrhs < 8) {
        printUsg();
        return;
    }
    
#define S1 prhs[0]
#define S2 prhs[1]
#define S3 prhs[2]
    
    if(!IS_REAL_2D(S1)) // check S1
        mexErrMsgTxt("S1 must be a real 2D matrix.");
    if(!IS_REAL_2D(S2)) // check S2
        mexErrMsgTxt("S2 must be a real 2D matrix.");
    if(!IS_REAL_2D(S3)) // check S3
        mexErrMsgTxt("S3 must be a real 2D matrix.");
    
    // init
    float* S1r = (float*)mxGetPr(S1);
    float* S2r = (float*)mxGetPr(S2);
    float* S3r = (float*)mxGetPr(S3);
    short  B   = static_cast<short>(mxGetM(S1)); // num of bins
    short  m   = static_cast<short>(mxGetN(S1))-1; // length of Seq1
    short  n   = static_cast<short>(mxGetN(S2))-1; // length of Seq2
    short  l   = static_cast<short>(mxGetN(S3))-1; // length of Seq3
    
    /*add 2015-4-6 */
    short* inValidSegs1 = (short *)mxGetPr(prhs[3]);
    short  n_invalid_1  = static_cast<short>(mxGetN(prhs[3]));
    
    short* inValidSegs2 = (short *)mxGetPr(prhs[4]);
    short  n_invalid_2  = static_cast<short>(mxGetN(prhs[4]));
    
    short* inValidSegs3 = (short *)mxGetPr(prhs[5]);
    short  n_invalid_3  = static_cast<short>(mxGetN(prhs[5]));
    /**/
    
    
    if (m != n || m != l) {
        mexErrMsgTxt("The input sequences must have the same length");
        return;
    }
    
    // Get parameters
    unsigned int maxIter;
    short        nRes;
    short        L, T;
    string       dist;
    bool         isVerbose;
    getParams(nrhs, prhs, L, T, maxIter, nRes, dist, isVerbose);
    
    Cset  R1, R2; // Cubic sets for splitting
    float lb, ub; // lower and upper bounds
    StateTri topS;   // top state
    priority_queue<StateTri, vector<StateTri>, CompareStateTri> Q; // priority queue
    priority_queue<StateTri, vector<StateTri>, CompareStateTri> Qtmp; // priority queue
    
    if (dist.compare("l1")!=0 && dist.compare("l2")!=0 && dist.compare("int")!=0) {
        mexPrintf("!!! We do not support \"%s\" distance. Please read the usage carefully!\n",dist.c_str());
        mexPrintf("!!! Please select from these distance options: [\"l1\",\"l2\",\"int\"]\n");
        return;
        
    }
    
    //Init output structure
    int          dimOut1[2] = {1,nRes};
    int          numFields1 = 3;
    const char*  fieldNames1[] = { "lo", "hi", "lb" };
    plhs[0] = mxCreateStructArray(2, dimOut1, numFields1, fieldNames1); // creates fields
    
    int          dimOut2[2] = {1,nRes};
    int          numFields2 = 1;
    const char*  fieldNames2[] = { "nIter" };
    plhs[1] = mxCreateStructArray(2, dimOut2, numFields2, fieldNames2);   // creates fields
    
    
    // Main loop below
    mexPrintf("===============================================\n");
    mexPrintf("+ Running TCD w/ %s distance\n", dist.c_str());
    mexPrintf("===============================================\n");
    for (short iRes=0; iRes<nRes; iRes++) {
        
        // init
        int nIter  = 0;      // counter for #iteration
        bool         isConv = false;  // flag for convergence
        
        Cset Rout = { {0,0,0,0,0,0}, {0,0,0,0,0,0}, false };
        
        // Update Q
        Cset R;
        if (iRes == 0) // discover 0 results so far
        {
            int LL = T + L; // L is minimal length
            if (isVerbose)
                mexPrintf("+ Initializing Q..\n");
            for (int t=1; t<=n-LL+1; t++)
            {
                // B1, B2
                R.lo[0] = t;      R.lo[2] = t;       R.lo[4] = t;
                R.hi[0] = T + t;  R.hi[2] = T + t;   R.hi[4] = T + t;
                
                // E1, E2
                R.lo[1] = L + t - 1;   R.lo[3] = L + t - 1;    R.lo[5] = L + t - 1;
                R.hi[1] = LL + t - 1;  R.hi[3] = LL + t - 1;   R.hi[5] = LL + t - 1;
                R.bValid = true;
                
                /*check collisions 2015-4-6*/
                check_conflict(R, inValidSegs1, n_invalid_1, inValidSegs2, n_invalid_2, inValidSegs3, n_invalid_3);
                
                if (R.isValid()){
                    getBounds(S1r, S2r, S3r, B, R, lb, ub, dist);
                    StateTri state = {lb,ub, R};
                    Q.push(state);
                }

            }
            
        } else { // discover more than one commonality
            Rout = topS.R;
            mexPrintf("+ Updating Q (size=%d -> ", Q.size());
            
            // Clean Qtmp
            while (!Qtmp.empty())
                Qtmp.pop();
            
            // Copy Q into Qtmp and clean Q
            while(!Q.empty()) {
                Qtmp.push(Q.top());
                Q.pop();
            }
            
            // Being to update Q
            StateTri tmpS;
            while(!Qtmp.empty()) {
                tmpS = Qtmp.top();
                Qtmp.pop();
                
                R = tmpS.R;
                // Update R
                updateR(topS.R, R);
                
                
                // Push {bound(R1), R1} to Q
                if (R.isValid()){
                    getBounds(S1r, S2r, S3r, B, R, lb, ub, dist);
                    StateTri St1 = {lb,ub, R};
                    Q.push(St1);
                }
                
            }
            mexPrintf("%d)\n", Q.size());
            if(Q.size() <= 0){
                break;
            }
            
            
            // Get topS
            topS = Q.top();
            R = topS.R;
            Q.pop();
        }
        
        
        // Get topS
        topS = Q.top();
        R = topS.R;
        Q.pop();
        
        //mexPrintf("TopS000: ");
        //topS.disp();
        
        if (isVerbose)
            mexPrintf("+ Running USD..\n");
        
        while (!isConv) {
            nIter++;
            
            // Check stopping criterion
            
            // diary out
            // mexPrintf("  # Iter%4d:\t",nIter);
            // topS.disp();
            
            if (nIter % 10000 == 0 && isVerbose) {
                mexPrintf("  # Iter%4d: ",nIter);
                topS.disp();
            }
            if (nIter >= maxIter) {
                mexPrintf("  # MaxIter%d reached: ", maxIter);
                topS.disp();
                isConv = true;
            }
            if (isUniqR(R)) {
                mexPrintf("  # Iter%4d: ",nIter);
                topS.disp();
                isConv = true;
                break;
            }
            
            // Split rect set into R1 and R2
            split(R, R1, R2);
            
            // Push R1 into queue if valid
            if (isvalid(R1, L)) {
                getBounds(S1r, S2r, S3r, B, R1, lb, ub, dist);
                StateTri S = {lb,ub,R1};
                Q.push(S);
                if (isVerbose) {
                    //mexPrintf("\t\tsplit R1: ");
                    //S.disp();
                }
            }
            
            // Push R2 into queue if valid
            if (isvalid(R2, L)) {
                getBounds(S1r, S2r, S3r, B, R2, lb, ub, dist);
                StateTri S = {lb,ub,R2};
                Q.push(S);
                if (isVerbose) {
                    //mexPrintf("\t\tsplit R2: ");
                    //S.disp();
                }
            }
            
            // Update R by top state
            topS = Q.top();
            R = topS.R;
            Q.pop();
            
        } // end of main loop
        
        // Assign output structure
        setField<short>(&plhs[0], iRes, "lo", 6, mxINT16_CLASS, topS.R.lo);    // set lo
        setField<short>(&plhs[0], iRes, "hi", 6, mxINT16_CLASS, topS.R.hi);    // set hi
        setField<float>(&plhs[0], iRes, "lb", 1, mxSINGLE_CLASS, &topS.lbound); // set lb
        setField<int>(&plhs[1], iRes, "nIter", 1, mxINT32_CLASS, &nIter); // set info.nIter
    } // end of iRes
    
    return;
}

// Set output struct with a specific type of data
template<typename T>
void setField(mxArray** A, const int structIdx, const char* fieldname, const int size, mxClassID mxType, const T* value) {
    mxArray *fieldValLo = mxCreateNumericMatrix(1, size, mxType, mxREAL);
    T* loVal = (T*)mxGetData(fieldValLo);
    memcpy(loVal, value, sizeof(T)*size);
    mxSetFieldByNumber(*A, structIdx, mxGetFieldNumber(*A,fieldname), fieldValLo);
}

// Get bounds according to different distance metrics
int getBounds(float* S1r, float* S2r, float* S3r, const short& B, const Cset& R, float& lb, float& ub, const string dist) {
    
    // Get |A+|
    float* Apv = new float[B];
    vSub(S1r + B*(R.hi[1]), S1r + B*(R.lo[0]-1), B, &Apv);
    float  Ap = vSum(Apv,B) + EPS; // |A+|
    
    // Get |B+|
    float* Bpv = new float[B];
    vSub(S2r + B*(R.hi[3]), S2r + B*(R.lo[2]-1), B, &Bpv);
    float Bp = vSum(Bpv, B) + EPS; // |B+|
    
    // Get |C+|
    float* Cpv = new float[B];
    vSub(S3r + B*(R.hi[5]), S3r + B*(R.lo[4]-1), B, &Cpv);
    float Cp = vSum(Cpv, B) + EPS; // |B+|
    
    
    // Get |A-|
    float  Am = EPS; // |A-|
    float* Amv = new float[B];
    if (R.lo[1] >= R.hi[0]) {
        vSub(S1r + B*(R.lo[1]), S1r + B*(R.hi[0]-1), B, &Amv);
        Am += vSum(Amv, B);
    } else {
        memset(Amv, 0, sizeof(float)*B);
    }
    
    // Get |B-|
    float  Bm = EPS;
    float* Bmv = new float[B];
    if (R.lo[3] >= R.hi[2]) {
        vSub(S2r + B*(R.lo[3]), S2r + B*(R.hi[2]-1), B, &Bmv);
        Bm += vSum(Bmv, B);
    } else {
        memset(Bmv, 0, sizeof(float)*B);
    }
    
    // Get |C-|
    float  Cm = EPS;
    float* Cmv = new float[B];
    if (R.lo[5] >= R.hi[4]) {
        vSub(S3r + B*(R.lo[5]), S3r + B*(R.hi[4]-1), B, &Cmv);
        Cm += vSum(Cmv, B);
    } else {
        memset(Cmv, 0, sizeof(float)*B);
    }
    
    lb = 0.0f;
    ub = 0.0f;
    short i;
    if (dist.compare("l1")==0) {
        for (i=0; i<B; i++) {
            lb += 2*max(0.0f, max(Amv[i]/Ap, max(Bmv[i]/Bp, Cmv[i]/Cp)) - min(Apv[i]/Am, min(Bpv[i]/Bm, Cpv[i]/Cm)));
            ub += 2*max(Apv[i]/Am, max(Bpv[i]/Bm, Cpv[i]/Cm)) - min(Amv[i]/Ap, min(Bmv[i]/Bp, Cmv[i]/Cp));
        }
        
    } else if (dist.compare("l2")==0) {
        for (i=0; i<B; i++) {
            lb += 3/2*pow(max(0.0f, max(Amv[i]/Ap, max(Bmv[i]/Bp,Cmv[i]/Cp)) - min(Apv[i]/Am, min(Bpv[i]/Bm,Cpv[i]/Cm))), 2.0f);
            ub += 2*pow(max(Apv[i]/Am, max(Bpv[i]/Bm, Cpv[i]/Cm)) - min(Amv[i]/Ap, min(Bmv[i]/Bp,Cmv[i]/Cp)), 2.0f);
        }
    }else if (dist.compare("int")==0) {
        for (i=0; i<B; i++) {
            lb -= min(Apv[i]/Am, min(Bpv[i]/Bm,Cpv[i]/Cm));
            ub -= min(Amv[i]/Ap, min(Bmv[i]/Bp,Cpv[i]/Cm));
        }
    }
    
    delete[] Apv;
    delete[] Bpv;
    delete[] Cpv;
    delete[] Amv;
    delete[] Bmv;
    delete[] Cmv;
    return 0;
}

// Vector substraction
void vSub(const float* v1, const float* v2, const short& B, float** out) {
    for (short i=0; i<B; i++)
        (*out)[i] = v1[i] - v2[i];
}

// Vector sum
float vSum(const float* v, const short& B) {
    float out = 0.0;
    for (short i=0; i<B; i++)
        out += v[i];
    return out;
}

// Check if a given R is unique
bool isUniqR(const Cset& R) {
    return (R.lo[0]==R.hi[0]) && (R.lo[1]==R.hi[1]) && (R.lo[2]==R.hi[2]) && (R.lo[3]==R.hi[3]) && (R.lo[4] == R.hi[4]) && (R.lo[5] == R.hi[5]);
}

// Check if a given rectangle set is valid
// In matlab: (R(1)<=R(4)-L) & (R(5)<=R(8)-L)
bool isvalid(const Cset& R, const short L) {
    return (R.lo[0]<=R.hi[1]-L+1) \
        && (R.lo[2]<=R.hi[3]-L+1) \
        && (R.lo[4]<=R.hi[5]-L+1);
}


// Rectify R if necessary
void rectify(Cset& R) {
    for (int i=0; i<6; i++)
        if (R.lo[i] > R.hi[i])
            swap(R.lo[i], R.hi[i]);
}

// Split a given rectangle R into two subsets R1 and R2 s.t. R=union(R1,R2)
void split(const Cset& R, Cset& R1, Cset& R2) {
    R1 = R;
    R2 = R;
    int ms = maxside(R);
    short sloc = (short)floor((float)(R1.lo[ms]+R1.hi[ms])/2); // split location
    R1.hi[ms] = sloc;
    R2.lo[ms] = sloc + 1;
    
    rectify(R1);
    rectify(R2);
}

// Find the maximal side in R
int maxside(const Cset& R) {
    short I[6]; // I for the 6 intervals B1,E1,B2,E2,B3,E3
    for (int i=0; i<6; i++)
        I[i] = R.hi[i] - R.lo[i];
    return (int)(distance(I, max_element(I, I+6)));
}

// Print the usage of TCD routine
void printUsg() {
    mexPrintf("  Usage: \n");
    mexPrintf("         [R,info] = tcd(S1,S2,L,[opt]);\n");
    mexPrintf("  Inputs:\n");
    mexPrintf("    S1,S2: integral images of vid1 and vid2 respectively (can be generated\n");
    mexPrintf("           by provided matlab script initHard.m or initSoft.m)\n");
    mexPrintf("    L:     minimal length allowed in TCD (\\ell in the paper)\n");
    mexPrintf("    opt:   options\n");
    mexPrintf("      .maxIter:   max #iter where TCD terminates [default:1e4]\n");
    mexPrintf("      .nRes:      number of resulting commonalties [default:1]\n");
    mexPrintf("      .dist:      metrics of interest (\"l1\",\"l2\",\"X2\",\"int\") [default:\"l1\"]\n");
    mexPrintf("      .isVerbose: verbose or not [default:true]\n");
    mexPrintf("  Outputs:\n");
    mexPrintf("    R:     discovered rectangle sets in the temporal search space\n");
    mexPrintf("    info:  some info output from the TCD algorithm\n\n");
}

// Update rectangle set to avoid collision
// Input R is split and output as R1 and R2
void updateR(Cset& rect, Cset& R) {
    
    short* B1lo = &R.lo[0]; short* B1hi = &R.hi[0];
    short* E1lo = &R.lo[1]; short* E1hi = &R.hi[1];
    short* B2lo = &R.lo[2]; short* B2hi = &R.hi[2];
    short* E2lo = &R.lo[3]; short* E2hi = &R.hi[3];
    short* B3lo = &R.lo[4]; short* B3hi = &R.hi[4];
    short* E3lo = &R.lo[5]; short* E3hi = &R.hi[5];
    
    short* b1 = &rect.lo[0]; short* e1 = &rect.lo[1];
    short* b2 = &rect.lo[2]; short* e2 = &rect.lo[3];
    short* b3 = &rect.lo[4]; short* e3 = &rect.lo[5];
    
    // no collision
    if (((*b1 > *E1hi) || (*e1 < *B1lo)) && \
        ((*b2 > *E2hi) || (*e2 < *B2lo)) && \
        ((*b3 > *E3hi) || (*e3 < *B3lo))){
        return;
    }
    
    //check collision dimension by dimension
    //check firt dimension
    if(*e1 <= *B1hi){
        *B1lo = *e1 + 1;
        if (*B1lo > *B1hi)
            R.setValid(false);
    }
    else if(*b1 >= *E1lo){
        *E1hi = *b1 - 1;
        if (*E1lo >= *E1hi)
            R.setValid(false);
    }
    else if((*b1 > *E1hi) || (*e1 < *B1lo)){
        ;
    }
    else{
        R.setValid(false);
    }
    
    //check second dimension
    if(*e2 <= *B2hi){
        *B2lo = *e2 + 1;
        if (*B2lo > *B2hi)
            R.setValid(false);
    }
    else if(*b2 >= *E2lo){
        *E2hi = *b2 -1;
        if (*E2lo > *E2hi)
            R.setValid(false);
    }
    else if((*b2 > *E2hi) || (*e2 < *B2lo)){
        ;
    }
    else{
        R.setValid(false);
    }
    
    //check third dimension
    if(*e3 <= *B3hi){
        *B3lo = *e3 + 1;
        if (*B3lo > *B3hi)
            R.setValid(false);
    }
    else if(*b3 >= *E3lo){
        *E3hi = *b3 - 1;
        if (*B3lo > *B3hi)
            R.setValid(false);
    }
    else if((*b3 > *E3hi) || (*e3 < *B3lo)){
        ;
    }
    else{
        R.setValid(false);
    }
    
}

// Copy contents for Cset
void copyR(Cset& Rin, Cset& Rout) {
    for (int i=0; i<6; i++) {
        Rout.lo[i] = Rin.lo[i];
        Rout.hi[i] = Rin.hi[i];
    }
    Rout.bValid = Rin.bValid;
}

// Copy contents for State
void copyS(StateTri& Sin, StateTri& Sout) {
    copyR(Sin.R, Sout.R);
    Sout.lbound = Sin.lbound;
    Sout.ubound = Sin.ubound;
}

// Check if alis within the interval
template<typename T> 
bool ismember(T a, T interval_begin, T interval_end) {
    return (a >= interval_begin) && (a <= interval_end);
}

// Check if a rectangle overlaps with a Cset
bool isOverlap(Cset& rect, Cset& R) {
    short* B1lo = &R.lo[0]; short* B1hi = &R.hi[0];
    short* E1lo = &R.lo[1]; short* E1hi = &R.hi[1];
    short* B2lo = &R.lo[2]; short* B2hi = &R.hi[2];
    short* E2lo = &R.lo[3]; short* E2hi = &R.hi[3];
    
    short* l = &rect.lo[0]; short* r = &rect.lo[1];
    short* t = &rect.lo[2]; short* b = &rect.lo[3];
    
    if (*r<*B1lo || *b<*B2lo || *l>*E1hi || *t>*E2hi || (*l>*B1hi && *t>*B2hi && *r<*E1lo && *b<*E2lo))
        return false;
    else
        return true;
}

void check_conflict(Cset& R, short *inValidSegs1, short n_invalid_1, short *inValidSegs2,short n_invalid_2, short *inValidSegs3, short n_invalid_3)
{
    short* B1lo = &R.lo[0]; short* B1hi = &R.hi[0];
    short* E1lo = &R.lo[1]; short* E1hi = &R.hi[1];
    short* B2lo = &R.lo[2]; short* B2hi = &R.hi[2];
    short* E2lo = &R.lo[3]; short* E2hi = &R.hi[3];
    short* B3lo = &R.lo[4]; short* B3hi = &R.hi[4];
    short* E3lo = &R.lo[5]; short* E3hi = &R.hi[5];
    
    // check B1lo, B1hi, E1lo, E1hi
    for(short i = 0; i < n_invalid_1; i ++)
    {
        short* b = &inValidSegs1[i*2]; short* e = &inValidSegs1[i*2+1];
        if ( *e < *B1lo || *b > *E1hi){
            continue;
        }else if(*e <= *B1hi){
            *B1lo = *e + 1;
        }else if(*b >= *E1lo){
            *E1hi = *b - 1;
        }else{
            R.setValid(false);
            return;
        }
        
        if (*B1lo > *B1hi){
            R.setValid(false);
            return;
        }
    }
    
    
    //check B2lo, B2hi, E2lo, E2hi
    for(short i = 0; i < n_invalid_2; i ++)
    {
        short* b = &inValidSegs2[i*2]; short* e = &inValidSegs2[i*2+1];
        if ( *e < *B2lo || *b > *E2hi){
            continue;
        }else if(*e <= *B2hi){
            *B2lo = *e + 1;
        }else if(*b >= *E2lo){
            *E2hi = *b - 1;
        }else{
            R.setValid(false);
            return;
        }
        
        if (*B2lo > *B2hi){
            R.setValid(false);
            return;
        }
    }
    
    //check B3lo, B3hi, E3lo, E3hi
    for(short i = 0; i < n_invalid_3; i ++)
    {
        short* b = &inValidSegs3[i*2]; short* e = &inValidSegs3[i*2+1];
        if ( *e < *B3lo || *b > *E3hi){
            continue;
        }else if(*e <= *B3hi){
            *B3lo = *e + 1;
        }else if(*b >= *E3lo){
            *E3hi = *b - 1;
        }else{
            R.setValid(false);
            return;
        }
        
        if (*B3lo > *B3hi){
            R.setValid(false);
            return;
        }
    }
}

/********************************
int getBounds(float* Snr,
              short* lengths,
              const short& nSeq,
              const short& B,
              const Cset& R,
              float &lb,
              float &ub,
              const string dist){  // NOT DEBUG YET
    
    float *pv = new float [nSeq * B];
    float *psum  = new float [nSeq];
    
    float *mv = new float [nSeq * B];
    float *msum = new float [nSeq];
    
    short *accLength = new short [nSeq-1];
    accLength[0] = 0;
    accLength[1] = lengths[0];
    for (int i = 2; i < nSeq-2; i ++){
        accLength[i] = lengths[i] + accLength[i-1];
    }
    
    
    for (int i = 0; i < nSeq; i ++){
        // Get |S^i+|
        float *Sir;
        Sir = Snr + B * accLength[i];
        vSub(Sir + B * R.hi[i*2+1], Sir + B*(R.lo[i*2]-1),B, &(pv+i*B));
        psum[i] = vSum(pv+i*B,B) + EPS;
        
        // Get |S^i-|
        msum[i] = EPS;
        if (R.lo[i*2+1] >= R.hi[i*2]) {
            vSub(Sir + B*(R.lo[i*2+1]), Sir + B*(R.hi[i*2]-1), B, &(mv+i*B));
            msum += vSum(mv + i*B, B);
        } else {
            memset(mv+i*B, 0, sizeof(float)*B);
        }
        
    }
    
    delete [] accLength;
    
    lb = 0.0f;
    ub = 0.0f;
    short b;
    for (b = 0; b < B; b++)
    {
        //compute h_b^i+/|S^i-|,  h_b^i-/|S^i+|
        float * hp = new float [nSeq];
        float * hm = new float [nSeq];
        for (short j = 0; j < nSeq; j++){
            hp[j] = *(pv + j*B + b)/msum[j];
            hm[j] = *(mv + j*B + b)/psum[j];
        }
        //sort hp, hm
        short * hp_order = new short [nSeq];
        short * hm_order = new short [nSeq];
        
        sort_vector(hp, hp_order, nSeq, 1);
        sort_vector(hm, hm_order, nSeq, 2);
        
        short vernier = 0;
        while (vernier < nSeq-1){
            float span = hm[hm_order[vernier]] - hp[hp_order[vernier]];
            if( span <= 0){
                break;
            }else{
                lb += span * (nSeq - vernier);
            }
            
            // remove hp_order[vernier] and hm_order[vernier]
            if (vernier+2 < nSeq){
                update_order(hp_order, nSeq, hm_order[vernier], vernier+2);
                update_order(hm_order, nSeq, hp_order[vernier], vernier+2);
            }
            
            vernier += 2;
        }
        
        
    }
    
    delete[] pv;
    delete[] mv;
    delete[] psum;
    delete[] msum;
    return 0;
}

void sort_vector(float *s, short* idx, short n, int t){
    // init idx
    float * tmps = new float [n];
    for (int i = 0; i < n; i ++){
        idx[i] = i;
        tmps[i] = s[i];
    }
    
    // bubble sort
    bool swapped = true;
    int j = 0;
    int tmp;
    while (swapped) {
        swapped = false;
        j++;
        for (int i = 0; i < n - j; i++) {
            if (arr[i] > arr[i + 1] && t == 1) { // ascend
                tmp = tmps[i];
                tmps[i] = tmps[i + 1];
                tmps[i + 1] = tmp;
                
                tmp = idx[i];
                idx[i] = idx[i + 1];
                idx[i + 1] = tmp;
                swapped = true;
            }
            else if(arr[i] < arr[i + 1] && t == 1) { //descend
                tmp = tmps[i];
                tmps[i] = tmps[i + 1];
                tmps[i + 1] = tmp;
                
                tmp = idx[i];
                idx[i] = idx[i + 1];
                idx[i + 1] = tmp;
                swapped = true;
            }
        }
    }
    
    delete [] tmps;
}

void update_order(short *s, short n, short rm_value, short bidx ){
    short tmp = 0;
    short rm_idx = 0;
    
    for (int i = bidx; i < n; i++ ){
        if (s[i] == rm_value){
            rm_idx = i;
            break;
        }
    }
    
    for (int i = rm_idx; i > bidx; i--)
        s[i] = s[i-1];
    s[bidx] = rm_value;
}

/*************************/
